module Main where

import Control.Monad  (sequence_)
import LevMar.Chart   (PlotValue, levmarChart)
import LevMar.Fitting
import NFunction      (($*))
import System.Random  (Random, mkStdGen, randoms)

import qualified SizedList as SL

-------------------------------------------------------------------------------
-- Some handy type level naturals

type N0 = Z
type N1 = S N0
type N2 = S N1
type N3 = S N2
type N4 = S N3
type N5 = S N4

-------------------------------------------------------------------------------
-- Model functions

constant  :: Num r => Model N1 r r
linear    :: Num r => Model N2 r r
quadratic :: Num r => Model N3 r r
cubic     :: Num r => Model N4 r r

constant        a _ = a
linear        b a x = b * x     + constant      a x
quadratic   c b a x = c * x*x   + linear      b a x
cubic     d c b a x = d * x*x*x + quadratic c b a x

trig  :: Floating r => Model N2 r r
trig a b x = a * cos x + b * sin x

-------------------------------------------------------------------------------
-- Jacobians

constantJacob  :: Num r => Jacobian N1 r r
linearJacob    :: Num r => Jacobian N2 r r
quadraticJacob :: Num r => Jacobian N3 r r
cubicJacob     :: Num r => Jacobian N4 r r

constantJacob        _ _ = 1     ::: Nil
linearJacob        _ a x = x     ::: constantJacob      a x
quadraticJacob   _ b a x = x*x   ::: linearJacob      b a x
cubicJacob     _ c b a x = x*x*x ::: quadraticJacob c b a x

trigJacob :: Floating r => Jacobian N2 r r
trigJacob _ _ x = cos x ::: sin x ::: Nil

-------------------------------------------------------------------------------
-- Test utility function

testChart :: ( ComposeN n
             , PlotValue r, Fractional r, LevMarable r, Random r
             , PlotValue a, Fractional a
             )
          => (Model n r a)
          -> Maybe (Jacobian n r a)
          -> SizedList n r
          -> [a]
          -> r
          -> IO ()
testChart f j ps xs noise = levmarChart f j
                                        initPs
                                        samples'
                                        1000
                                        opts
                                        Nothing
                                        Nothing
                                        noLinearConstraints
                                        Nothing
                           >> return ()
    where
      ns         = take (length xs) $ randoms $ mkStdGen rndGenSeed
      initPs     = (SL.replicate (SL.length ps) 0)
      samples    = zip xs $ map (f $* ps) xs
      samples'   = zipWith (\(x, y) n -> (x, y + (n - 0.5) * 2 * noise)) samples ns
      rndGenSeed = 123456
      opts       = defaultOpts { optStopNormInfJacTe = 1e-15
                               , optStopNorm2Dp      = 1e-15
                               , optStopNorm2E       = 1e-20
                               }

-------------------------------------------------------------------------------
-- Examples

testConstant :: IO ()
testConstant = testChart constant (Just constantJacob)
                         (42 ::: Nil)
                         [-100 .. 100]
                         (10 :: Double)

testLinear :: IO ()
testLinear = testChart linear (Just linearJacob)
                       (3.14 ::: -8 ::: Nil)
                       [-100 .. 100]
                       (300 :: Double)

testQuadratic :: IO ()
testQuadratic = testChart quadratic (Just quadraticJacob)
                          (0.2 ::: 8 ::: 23 ::: Nil)
                          [-100 .. 100]
                          (1000 :: Double)

testCubic :: IO ()
testCubic = testChart cubic (Just cubicJacob)
                      (-0.05 ::: 0.5 ::: -12 ::: 10 ::: Nil)
                      [-100 .. 100]
                      (8000 :: Double)

testTrig :: IO ()
testTrig = testChart trig (Just trigJacob)
                     (100 ::: 102 ::: Nil)
                     (map (/4) [0..400])
                     (50 :: Double)

-------------------------------------------------------------------------------
-- Main

main :: IO ()
main = sequence_ [ testConstant
                 , testLinear
                 , testQuadratic
                 , testCubic
                 , testTrig
                 ]
